/*
 * pseudo.cpp
 * Implementation of the PsuedoSprite class.
 *
 * Copyright 2006 by Dale McCoy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<sstream>
#include<iostream>
#include<iomanip>
#include<cstdarg>

using namespace std;

#include"renum.h"
#include"pseudo.h"
#include"globals.h"
#include"messages.h"
#include"inlines.h"
#include"command.h"

extern int NFOversion;
#define SetVersion(x)\
    (NFOversion=max((x),NFOversion))


enum{HEX,TEXT,UTF8,ENDQUOTE,NOBREAK};


PseudoSprite::PseudoSprite(const string&sprite):orig(sprite),valid(true){
    istringstream in(sprite);
    ostringstream out;
    char ch;
    bool literal=false;
    while(in.get(ch)){

//This section is a rewrite of Petr Baudis's (Pasky's) patch to info.cc:
//http://pasky.or.cz/~pasky/dev/openttd/grfcodec/grfcodec+unix2.patch
// now incorporated into GRFCodec 0.9.6 and later
//That patch is Copyright Petr Baudis
        if(ch=='"'){
            if(out.str().length()%2){
                IssueMessage(0,OFFSET_STRING);
                Invalidate();
                return;
            }
            SetVersion(5);
            literal=!literal;
            continue;
        }
        if(literal){
            /*TODO: Uncomment if/when GRFCodec DTRT.
            if(ch=='\\'){
                in.get(ch);
                if(ch=='\n')continue;
                if(ch=='x'||ch=='X'){
                    in.get(ch);
                    out.put(ch);
                    in.get(ch);
                    out.put(ch);
                    continue;
                }//*/
            out<<itoa((unsigned char)ch,16,2);
            continue;
        }
        if(COMMENT.find(ch)!=NPOS){
            in.ignore(INT_MAX,'\n');
            continue;
        }
        if(isspace(ch))continue;
        if(string(VALID_PSEUDO).find(ch)==NPOS){
            IssueMessage(0,INVALID_CHARACTER,ch);
            Invalidate();
            return;
        }
        out<<(char)toupper(ch);
    }
    if(literal){
        IssueMessage(0,UNTERMINATED_STRING);
        Invalidate();
        return;
    }
    if((packed=out.str()).length()%2){
        IssueMessage(0,BAD_SIZE);
        Invalidate();
        return;
    }
}

bool PseudoSprite::MayBeSprite(const string&sprite){
    istringstream in(sprite);
    char ch;
    while(in.get(ch)){
        if(ch=='"')return true;
        if(COMMENT.find(ch)!=NPOS){
            in.ignore(INT_MAX,'\n');
            continue;
        }
        if(isspace(ch)||string(VALID_PSEUDO).find(ch)==NPOS)continue;
        return true;
    }
    return false;
}

uint PseudoSprite::Length()const{
    return(uint)(valid?packed.length()/2:0);
}

static uint(PseudoSprite::* const ExtractFuncs[])(uint)const=
    {NULL,&PseudoSprite::ExtractByte,&PseudoSprite::ExtractWord,&PseudoSprite::ExtractExtended,&PseudoSprite::ExtractDword};

uint PseudoSprite::ExtractVariable(uint offs,uint length)const{
    VERIFY((length&&length<5),PseudoSprite::ExtractVariable,length);
    return(this->*(ExtractFuncs[length]))(offs);
}

uint PseudoSprite::ExtendedLen(uint offs)const{
    return ExtractByte(offs)!=0xFF?1:3;
}

uint PseudoSprite::ExtractByte(uint offs)const{
    VERIFY(valid,PseudoSprite::ExtractByte,valid);
    if(Length()<=offs)
        throw offs|1<<24;
    return (ctoi(packed[2*offs])<<4|ctoi(packed[2*offs+1]));
}

uint PseudoSprite::ExtractWord(uint offs)const{
    try{
        return ExtractByte(offs)|ExtractByte(offs+1)<<8;
    }catch(unsigned int){
        throw offs|2<<24;
    }
}

uint PseudoSprite::ExtractExtended(uint offs)const{
    uint byte=ExtractByte(offs);
    if(byte!=0xFF)return byte;
    return ExtractWord(offs+1);
}

uint PseudoSprite::ExtractDword(uint offs)const{
    try{
        return ExtractWord(offs)|ExtractWord(offs+2)<<16;
    }catch(unsigned int){
        throw offs|4<<24;
    }
}

ostream&operator<<(ostream&out,const PseudoSprite&sprite){
    sprite.output(out);
    return out;
}

ostream&PseudoSprite::output(ostream&out)const{
    if(!valid){
        istringstream datastream(orig);
        string line;
        getline(datastream,line);
        out<<COMMENT_PREFIX<<"    0 * 0\t "<<line<<endl;
        while(getline(datastream,line)){
            if(!is_comment(line))out<<COMMENT_PREFIX;
            out<<line<<endl;
        }
        return out;
    }
    if(Length())
        out<<setw(5)<<spritenum()<<" * "<<(GetState(DIFF)?0:Length())<<"\t ";
    return out<<orig;
}

void PseudoSprite::Invalidate(){
    IssueMessage(0,PARSE_FAILURE,_spritenum);
    valid=false;
    SetCode(EPARSE);
}
